from __future__ import annotations

import warnings
from collections import defaultdict
from collections.abc import Callable
from datetime import date, datetime, time, timedelta
from enum import Enum
from typing import Any, Optional, TypeVar
from urllib.parse import unquote_plus
from uuid import UUID

from kiota_abstractions.date_utils import (
    parse_timedelta_string, datetime_from_iso_format_compat, time_from_iso_format_compat
)
from kiota_abstractions.serialization import Parsable, ParsableFactory, ParseNode

T = TypeVar("T", bool, str, int, float, UUID, datetime, timedelta, date, time, bytes)

U = TypeVar("U", bound=Parsable)

K = TypeVar("K", bound=Enum)


class FormParseNode(ParseNode):
    """Represents a parse node that can be used to parse a form url encoded string."""

    def __init__(self, raw_value: str) -> None:
        self._raw_value = raw_value
        self._node = unquote_plus(raw_value)
        self._fields = self._get_fields(raw_value)
        self._on_before_assign_field_values: Optional[Callable[[Parsable], None]] = None
        self._on_after_assign_field_values: Optional[Callable[[Parsable], None]] = None

    def get_str_value(self) -> Optional[str]:
        """Gets the string value from the node
        Returns:
            str: The string value of the node
        """
        if self._node and self._node != "null":
            try:
                return str(self._node)
            except:
                return None
        return None

    def get_bool_value(self) -> Optional[bool]:
        """Gets the boolean value of the node
        Returns:
            bool: The boolean value of the node
        """
        if self._node and self._node != "null":
            return self._node.lower() == "true"
        return None

    def get_int_value(self) -> Optional[int]:
        """Gets the integer value of the node
        Returns:
            int: The integer value of the node
        """
        if self._node and self._node != "null":
            try:
                return int(self._node)
            except:
                return None
        return None

    def get_float_value(self) -> Optional[float]:
        """Gets the float value of the node
        Returns:
            float: The integer value of the node
        """
        if self._node and self._node != "null":
            try:
                return float(self._node)
            except:
                return None
        return None

    def get_uuid_value(self) -> Optional[UUID]:
        """Gets the UUID value of the node
        Returns:
            UUID: The GUID value of the node
        """
        if self._node and self._node != "null":
            try:
                return UUID(self._node)
            except:
                return None
        return None

    def get_datetime_value(self) -> Optional[datetime]:
        """Gets the datetime value of the node
        Returns:
            datetime: The datetime value of the node
        """
        if self._node and self._node != "null":
            try:
                return datetime_from_iso_format_compat(self._node)
            except:
                return None
        return None

    def get_timedelta_value(self) -> Optional[timedelta]:
        """Gets the timedelta value of the node
        Returns:
            timedelta: The timedelta value of the node
        """
        if self._node and self._node != "null":
            try:
                return parse_timedelta_string(self._node)
            except:
                return None
        return None

    def get_date_value(self) -> Optional[date]:
        """Gets the date value of the node
        Returns:
            date: The datevalue of the node in terms on year, month, and day.
        """
        if self._node and self._node != "null":
            try:
                return date.fromisoformat(self._node)
            except:
                return None
        return None

    def get_time_value(self) -> Optional[time]:
        """Gets the time value of the node
        Returns:
            time: The time value of the node in terms of hour, minute, and second.
        """
        if self._node and self._node != "null":
            try:
                return time_from_iso_format_compat(self._node)
            except:
                return None
        return None

    def get_bytes_value(self) -> Optional[bytes]:
        """Get the bytes value of the node
        Returns:
            bytes: The decoded bytes value
        """
        if self._node and self._node != "null":
            try:
                base64_string = str(self._node)
                return base64_string.encode("utf-8")
            except:
                return None
        return None

    def get_child_node(self, field_name: str) -> Optional[ParseNode]:
        """Gets the child node of the node
        Returns:
            Optional[ParseNode]: The child node of the node
        """
        if field_name in self._fields:
            return FormParseNode(self._fields[field_name])
        return None

    def get_collection_of_primitive_values(self, primitive_type: type[T]) -> Optional[list[T]]:
        """Gets the collection of primitive values of the node
        Args:
            primitive_type: The type of primitive to return.
        Returns:
            list[T]: The collection of primitive values
        """
        if not primitive_type:
            raise Exception("Primitive type for deserialization cannot be null")

        primitive_types = {bool, str, int, float, UUID, datetime, timedelta, date, time, bytes}
        if primitive_type in primitive_types:
            items = self._node.split(',')
            result: list[T] = []
            for item in items:
                current_parse_node = self._create_new_node(item)
                method_name = f"get_{primitive_type.__name__.lower()}_value"
                method = getattr(current_parse_node, method_name)
                result.append(method())
            return result
        raise Exception(f"Encountered an unknown type during deserialization {primitive_type}")

    def get_collection_of_object_values(self, factory: ParsableFactory[U]) -> Optional[list[U]]:
        raise Exception("Collection of object values is not supported with uri form encoding.")

    def get_collection_of_enum_values(self, enum_class: K) -> Optional[list[K]]:
        """Gets the collection of enum values of the node
        Returns:
            list[K]: The collection of enum values
        """
        values = self._node.split(',')
        if values:
            return list(
                map(
                    lambda x: self._create_new_node(x).get_enum_value(enum_class),  # type: ignore
                    values
                )
            )
        return []

    def get_enum_value(self, enum_class: K) -> Optional[K]:
        """Gets the enum value of the node
        Returns:
            Optional[K]: The enum value of the node
        """

        if not self._node:
            return None
        enum_values = [e.value for e in enum_class]  # type: ignore
        if self._node in enum_values:
            return enum_class(self._node)  # type: ignore
        values = self._node.split(',')
        if not len(values) > 1:
            raise Exception(f'Invalid value: {self._node} for enum {enum_class}.')
        result = []
        for value in values:
            if value not in enum_values:
                raise Exception(f'Invalid value: {value} for enum {enum_class}.')
            result.append(enum_class(value))  # type: ignore
        return result  # type: ignore

    def get_object_value(self, factory: ParsableFactory[U]) -> U:
        """Gets the model object value of the node
        Returns:
            Parsable: The model object value of the node
        """

        result = factory.create_from_discriminator_value(self)
        if on_before := self.on_before_assign_field_values:
            on_before(result)
        self._assign_field_values(result)
        if on_after := self.on_after_assign_field_values:
            on_after(result)
        return result

    @property
    def on_before_assign_field_values(self) -> Optional[Callable[[Parsable], None]]:
        """Gets the callback called before the node is deserialized.
        Returns:
            Callable[[Parsable], None]: the callback called before the node is deserialized.
        """
        return self._on_before_assign_field_values

    @on_before_assign_field_values.setter
    def on_before_assign_field_values(self, value: Callable[[Parsable], None]) -> None:
        """Sets the callback called before the node is deserialized.
        Args:
            value (Callable[[Parsable], None]): the callback called before the node is
            deserialized.
        """
        self._on_before_assign_field_values = value

    @property
    def on_after_assign_field_values(self) -> Optional[Callable[[Parsable], None]]:
        """Gets the callback called before the node is deserialized.
        Returns:
            Callable[[Parsable], None]: the callback called before the node is deserialized.
        """
        return self._on_after_assign_field_values

    @on_after_assign_field_values.setter
    def on_after_assign_field_values(self, value: Callable[[Parsable], None]) -> None:
        """Sets the callback called after the node is deserialized.
        Args:
            value (Callable[[Parsable], None]): the callback called after the node is
            deserialized.
        """
        self._on_after_assign_field_values = value

    def _assign_field_values(self, item: Parsable) -> None:
        """Assigns the field values to the model object"""

        # if object is null
        if not self._fields:
            return

        item_additional_data = None
        if not hasattr(item, "additional_data") or item.additional_data is None:
            item_additional_data = {}
        else:
            item_additional_data = item.additional_data

        field_deserializers = item.get_field_deserializers()

        for field_name, field_value in self._fields.items():
            if field_name in field_deserializers:
                if field_value is None:
                    continue
                field_deserializer = field_deserializers[field_name]
                field_deserializer(FormParseNode(field_value))
            elif item_additional_data is not None:
                item_additional_data[field_name] = self.try_get_anything(field_value)
            else:
                warnings.warn(
                    f"Found additional property {field_name} to \
                    deserialize but the model doesn't support additional data"
                )

    def try_get_anything(self, value: Any) -> Any:
        if isinstance(value, (int, float, bool)) or value is None:
            return value
        if isinstance(value, list):
            return list(map(self.try_get_anything, value))
        if isinstance(value, dict):
            return dict(map(lambda x: (x[0], self.try_get_anything(x[1])), value.items()))
        if isinstance(value, str):
            try:
                datetime_obj = datetime_from_iso_format_compat(value)
                return datetime_obj
            except ValueError:
                pass
            try:
                return UUID(value)
            except ValueError:
                pass
            try:
                return parse_timedelta_string(value)
            except ValueError:
                pass
            try:
                return date.fromisoformat(value)
            except ValueError:
                pass
            try:
                return time_from_iso_format_compat(value)
            except ValueError:
                pass
            return value
        raise ValueError(f"Unexpected additional value type {type(value)} during deserialization.")

    def _create_new_node(self, node: Any) -> FormParseNode:
        new_node: FormParseNode = FormParseNode(node)
        new_node.on_before_assign_field_values = self.on_before_assign_field_values
        new_node.on_after_assign_field_values = self.on_after_assign_field_values
        return new_node

    def _get_fields(self, raw_value: str) -> dict[str, str]:
        fields = raw_value.split('&')
        field_values = defaultdict(list)
        for field in fields:
            if '=' in field:
                key, value = field.split('=', 1)
                key = self._sanitize_key(key)
                value = value.strip()
                field_values[key].append(value)

        # Convert lists to comma-separated strings
        result: dict[str, str] = {}
        for key in field_values:
            result[key] = ','.join(field_values[key])
        return result

    def _sanitize_key(self, key: str) -> str:
        if not key:
            return key
        return unquote_plus(key.strip())



from __future__ import annotations

import base64
from collections.abc import Callable
from datetime import date, datetime, time, timedelta
from enum import Enum
from typing import Any, Optional, TypeVar
from urllib.parse import quote_plus
from uuid import UUID

from kiota_abstractions.serialization import Parsable, SerializationWriter

T = TypeVar("T")
U = TypeVar("U", bound=Parsable)
K = TypeVar("K", bound=Enum)


class FormSerializationWriter(SerializationWriter):

    def __init__(self) -> None:
        self.writer: str = ""
        self.depth = 0

        self._on_start_object_serialization: Optional[Callable[[Parsable, SerializationWriter],
                                                               None]] = None
        self._on_before_object_serialization: Optional[Callable[[Parsable], None]] = None
        self._on_after_object_serialization: Optional[Callable[[Parsable], None]] = None

    def write_str_value(self, key: Optional[str], value: Optional[str]) -> None:
        """Writes the specified string value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Optional[str]): The string value to be written.
        """
        if key and value:
            if len(self.writer) > 0:
                self.writer += "&"
            self.writer += f"{quote_plus(key.strip())}={quote_plus(value.strip())}"

    def write_bool_value(self, key: Optional[str], value: Optional[bool]) -> None:
        """Writes the specified boolean value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Optional[bool]): The boolean value to be written.
        """
        if key and (value or value is False):
            self.write_str_value(key, str(value).lower())

    def write_int_value(self, key: Optional[str], value: Optional[int]) -> None:
        """Writes the specified integer value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Optional[int]): The integer value to be written.
        """
        if key and (value or value == 0):
            self.write_str_value(key, str(value))

    def write_float_value(self, key: Optional[str], value: Optional[float]) -> None:
        """Writes the specified float value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Optional[float]): The float value to be written.
        """
        if key and (value or value == 0):
            self.write_str_value(key, str(value))

    def write_uuid_value(self, key: Optional[str], value: Optional[UUID]) -> None:
        """Writes the specified uuid value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Optional[UUId]): The uuid value to be written.
        """
        if key and value:
            self.write_str_value(key, str(value))

    def write_datetime_value(self, key: Optional[str], value: Optional[datetime]) -> None:
        """Writes the specified datetime offset value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Optional[datetime]): The datetime offset value to be written.
        """
        if key and value:
            if isinstance(value, datetime):
                self.write_str_value(key, str(value.isoformat()))
            else:
                self.write_str_value(key, str(value))

    def write_timedelta_value(self, key: Optional[str], value: Optional[timedelta]) -> None:
        """Writes the specified timedelta value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Optional[timedelta]): The timedelta value to be written.
        """
        if key and value:
            self.write_str_value(key, str(value))

    def write_date_value(self, key: Optional[str], value: Optional[date]) -> None:
        """Writes the specified date value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Optional[date]): The date value to be written.
        """
        if key and value:
            self.write_str_value(key, str(value))

    def write_time_value(self, key: Optional[str], value: Optional[time]) -> None:
        """Writes the specified time value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Optional[time]): The time value to be written.
        """
        if key and value:
            self.write_str_value(key, str(value))

    def write_bytes_value(self, key: Optional[str], value: Optional[bytes]) -> None:
        """Writes the specified byte array as a base64 string to the stream with an optional
        given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (bytes): The byte array to be written.
        """
        if key and isinstance(value, bytes):
            base64_bytes = base64.b64encode(value)
            base64_string = base64_bytes.decode('utf-8')
            self.write_str_value(key, base64_string)

    def write_collection_of_primitive_values(
        self, key: Optional[str], values: Optional[list[T]]
    ) -> None:
        """Writes the specified collection of primitive values to the stream with an optional
        given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            values (Optional[list[T]]): The collection of primitive values to be written.
        """
        primitive_types = [bool, str, int, float, UUID, datetime, timedelta, date, time, Enum]
        if key and values:
            for val in values:
                if type(val) in primitive_types:
                    method = getattr(self, f'write_{type(val).__name__.lower()}_value')
                    method(key, val)

    def write_collection_of_enum_values(
        self, key: Optional[str], values: Optional[list[K]]
    ) -> None:
        """Writes the specified collection of enum values to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            values Optional[list[K]): The enum values to be written.
        """
        if key and values:
            if isinstance(values, list):
                for val in values:
                    if isinstance(val, Enum):
                        self.write_str_value(key, str(val.value))

    def write_enum_value(self, key: Optional[str], value: Optional[K]) -> None:
        """Writes the specified enum value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Optional[K]): The enum value to be written.
        """
        if key and value:
            if isinstance(value, Enum):
                self.write_str_value(key, str(value.value))
            if isinstance(value, list):
                values = ",".join([str(val.value) for val in value])
                self.write_str_value(key, values)

    def write_collection_of_object_values(
        self, key: Optional[str], values: Optional[list[U]]
    ) -> None:
        """Writes the specified collection of model objects to the stream with an optional
        given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            values (Optional[list[U]]): The collection of model objects to be written.
        """
        raise Exception("Form serialization does not support collections.")

    def write_object_value(
        self, key: Optional[str], value: Optional[U], *additional_values_to_merge: Optional[U]
    ) -> None:
        """Writes the specified model object to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Parsable): The model object to be written.
            additional_values_to_merge (tuple[Parsable]): The additional values to merge to the
            main value when serializing an intersection wrapper.
        """
        if self.depth > 0:
            raise Exception("Form serialization does not support nested objects.")
        self.depth += 1
        temp_writer = self._create_new_writer()

        if value is not None:
            self._serialize_value(temp_writer, value)

        if additional_values_to_merge:
            for additional_value in filter(lambda x: x is not None, additional_values_to_merge):
                self._serialize_value(temp_writer, additional_value)  # type: ignore
                if on_after := self.on_after_object_serialization:
                    on_after(additional_value)  # type: ignore

        if value and self._on_after_object_serialization:
            self._on_after_object_serialization(value)

        if len(self.writer) > 0:
            self.writer += "&"
        self.writer += f"{quote_plus(key.strip()) if key is not None else ''}={temp_writer.writer}"
        self.depth -= 1

    def write_null_value(self, key: Optional[str]) -> None:
        """Writes a null value for the specified key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
        """
        if key:
            self.write_str_value(key, "null")

    def write_additional_data_value(self, value: dict[str, Any]) -> None:
        """Writes the specified additional data to the stream.
        Args:
            value (dict[str, Any]): he additional data to be written.
        """
        if isinstance(value, dict):
            for key, val in value.items():
                if isinstance(val, Parsable):
                    raise Exception("Form serialization does not support nested objects")
                self.write_any_value(key, val)

    def get_serialized_content(self) -> bytes:
        """Gets the value of the serialized content.
        Returns:
            bytes: The value of the serialized content.
        """
        if self.writer:
            return self.writer.encode('utf-8')
        return b''

    @property
    def on_before_object_serialization(self) -> Optional[Callable[[Parsable], None]]:
        """Gets the callback called before the object gets serialized.
        Returns:
            Optional[Callable[[Parsable], None]]:the callback called before the object
            gets serialized.
        """
        return self._on_before_object_serialization

    @on_before_object_serialization.setter
    def on_before_object_serialization(self, value: Optional[Callable[[Parsable], None]]) -> None:
        """Sets the callback called before the objects gets serialized.
        Args:
            value (Optional[Callable[[Parsable], None]]): the callback called before the objects
            gets serialized.
        """
        self._on_before_object_serialization = value

    @property
    def on_after_object_serialization(self) -> Optional[Callable[[Parsable], None]]:
        """Gets the callback called after the object gets serialized.
        Returns:
            Optional[Optional[Callable[[Parsable], None]]]: the callback called after the object
            gets serialized.
        """
        return self._on_after_object_serialization

    @on_after_object_serialization.setter
    def on_after_object_serialization(self, value: Optional[Callable[[Parsable], None]]) -> None:
        """Sets the callback called after the objects gets serialized.
        Args:
            value (Optional[Callable[[Parsable], None]]): the callback called after the objects
            gets serialized.
        """
        self._on_after_object_serialization = value

    @property
    def on_start_object_serialization(
        self
    ) -> Optional[Callable[[Parsable, SerializationWriter], None]]:
        """Gets the callback called right after the serialization process starts.
        Returns:
            Optional[Callable[[Parsable, SerializationWriter], None]]: the callback called
            right after the serialization process starts.
        """
        return self._on_start_object_serialization

    @on_start_object_serialization.setter
    def on_start_object_serialization(
        self, value: Optional[Callable[[Parsable, SerializationWriter], None]]
    ) -> None:
        """Sets the callback called right after the serialization process starts.
        Args:
            value (Optional[Callable[[Parsable, SerializationWriter], None]]): the callback
            called right after the serialization process starts.
        """
        self._on_start_object_serialization = value

    def write_non_parsable_object_value(self, key: Optional[str], value: T) -> None:
        """Writes the specified value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (object): The value to be written.
        """
        if key and value:
            if hasattr(value, '__dict__'):
                temp_writer = self._create_new_writer()
                for k, v in value.__dict__.items():
                    temp_writer.write_any_value(k, v)
                if len(self.writer) > 0:
                    self.writer += "&"
                self.writer += f"{quote_plus(key.strip())}={temp_writer.writer}"

    def write_any_value(self, key: Optional[str], value: Any) -> Any:
        """Writes the specified value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value Any): The value to be written.
        """
        primitive_types = [bool, str, int, float, UUID, datetime, timedelta, date, time, Enum]
        if key and value:
            value_type = type(value)
            if value_type in primitive_types:
                method = getattr(self, f'write_{value_type.__name__.lower()}_value')
                method(key, value)
            elif isinstance(value, list):
                if all(isinstance(x, Enum) for x in value):
                    self.write_collection_of_enum_values(key, value)
                else:
                    self.write_collection_of_primitive_values(key, value)
            elif isinstance(value, Parsable):
                self.write_object_value(key, value)
            elif hasattr(value, '__dict__'):
                self.write_non_parsable_object_value(key, value)
            else:
                raise TypeError(
                    f"Encountered an unknown type during serialization {value_type} \
                        with key {key}"
                )

    def _serialize_value(self, temp_writer: FormSerializationWriter, value: U):
        if on_before := self.on_before_object_serialization:
            on_before(value)
        if on_start := self.on_start_object_serialization:
            on_start(value, self)

        value.serialize(temp_writer)

    def _create_new_writer(self) -> FormSerializationWriter:
        writer = FormSerializationWriter()
        writer.on_before_object_serialization = self.on_before_object_serialization
        writer.on_after_object_serialization = self.on_after_object_serialization
        writer.on_start_object_serialization = self.on_start_object_serialization
        return writer
